from fns.cluster import *
from fns.constants import *
from fns.dataframe import *
from fns.decorators import *
from fns.exceptions import *
from fns.json_encoders import *
from fns.metrics import *
from fns.model_selection import *
from fns.preprocessing import *
from fns.streamlit_utils import *
from fns.text import *
from fns.vision import *
from fns.fns import *
from fns.plot import *
from sklearn.dummy import DummyRegressor, DummyClassifier
from sklearn.tree import DecisionTreeClassifier, DecisionTreeRegressor
from sklearn.linear_model import LogisticRegression, LinearRegression
from sklearn.naive_bayes import GaussianNB, BernoulliNB, MultinomialNB
from sklearn.ensemble import RandomForestClassifier, RandomForestRegressor
from sklearn.cluster import KMeans, AffinityPropagation, AgglomerativeClustering
from sklearn.decomposition import PCA, TruncatedSVD, NMF
from sklearn.manifold import TSNE
from sklearn.model_selection import train_test_split, GridSearchCV, RandomizedSearchCV, cross_val_score
from sklearn.feature_extraction.text import TfidfVectorizer, CountVectorizer, HashingVectorizer
from sklearn.preprocessing import normalize, scale, minmax_scale
from sklearn.metrics import accuracy_score, precision_score, recall_score, confusion_matrix
from sklearn.metrics.pairwise import cosine_distances, cosine_similarity, euclidean_distances
import matplotlib.pyplot as plt
from itertools import chain, product, permutations, combinations, islice
from functools import lru_cache
from collections import Counter, OrderedDict, defaultdict
from pathlib import Path
import math

try:
    import torch
    import torch.optim as optim
    import torch.nn as nn
    from torch.utils.data import Dataset, TensorDataset, DataLoader
except ImportError:
    pass



import time
from typing import Callable


def timeit(func: Callable) -> Callable:
    """
    Decorator to calculate time taken for a function to complete.

    Args:
        func: Python Function

    Returns:
        Decorated function
    """
    start_time = time.time()

    def inner(*args, **kwargs):
        func(*args, **kwargs)
        total_time_taken = time.time() - start_time
        print('Total time taken: {} seconds'.format(total_time_taken))

    return inner


def show_shapes(fxn: Callable) -> Callable:
    """
    Decorator to log dataframe shape before and after applying a function.

    Args:
        fxn: Function that takes a dataframe as argument

    Returns:
        function
    """

    def inner(df):
        print(f'Shape before {fxn.__name__}', df.shape)
        out_df = fxn(df)
        print(f'Shape after {fxn.__name__}', out_df.shape)
        return out_df

    return inner


def deduplicate(fxn: Callable) -> Callable:
    """
    Decorator to deduplicate results of a function.

    Usage:
    ```python
    @deduplicate
    def test():
        return [1, 2, 3, 1]
    ```

    Args:
        fxn: Function

    Returns:
        Function
    """

    def inner(*args, **kwargs):
        return list(set(fxn(*args, **kwargs)))

    return inner



import math
from collections import Counter
from typing import Dict, Callable, List

import numpy as np
import timeit


def benchmark_function(fn: Callable,
                       repeat: int = 5) -> Dict:
    """
    Benchmark time taken for a function and return metrics.

    Args:
        fn: A python function
        repeat: Number of samples

    Returns:
        Dictionary of total times, mean and std of times
    """
    iteration_times = timeit.repeat(fn,
                                    repeat=repeat,
                                    number=1)
    return {'time': iteration_times,
            'mean': np.mean(iteration_times),
            'std': np.std(iteration_times)}


def baseline_accuracy(labels: List) -> float:
    """
    Get accuracy for always majority class classifier.

    Usage:
    ```python
    >>> baseline_accuracy([0, 1])
    50.0
    ```

    Args:
        labels: List of class labels.

    Returns:
        Baseline accuracy
    """
    (label, count), *_ = Counter(labels).most_common(1)
    return count / len(labels) * 100.0


def missing_value_percent(df):
    """
    Get the percentage of missing values in each column.

    Args:
        df: Pandas DataFrame

    Returns:
        Percentage of missing value in each column.
    """
    num_rows = len(df)
    return (df.isna().sum() / num_rows * 100.0).sort_values(ascending=False)


def na_percent(df):
    return missing_value_percent(df)


def n_clusters(data) -> int:
    """
    Generate number of clusters to create.

    Heuristic:
    Number of clusters = square root of total data points

    Args:
        data: Total number of data points or the data point itself

    Returns:
        Number of clusters
    """
    if type(data) is int:
        total_rows = data
    else:
        total_rows = len(set(data))
    return int(math.sqrt(total_rows))


def vector_similarity(u, v) -> float:
    angle = np.dot(u, v) / (np.linalg.norm(u) * np.linalg.norm(v))
    return float(angle)


def jaccard(x, y) -> float:
    """
    Compute jaccard similarity (intersection over union).

    Args:
        x: Array-like object
        y: Array-like object

    Returns:
        Intersection Over Union score
    """
    s1 = set(x)
    s2 = set(y)
    if len(s1) == 0 and len(s2) == 0:
        return 0
    return len(s1 & s2) / len(s1 | s2)



from sklearn.metrics import ConfusionMatrixDisplay, confusion_matrix


def confusion_matrix_plot(y_true, y_pred):
    cm = confusion_matrix(y_true, y_pred)
    plot = ConfusionMatrixDisplay(confusion_matrix=cm).plot()
    plot.ax_.set_title('Confusion Matrix')



from functools import lru_cache
import pandas as pd
import numpy as np

NEGATION_WORDS = ['no', 'not', 'never', 'none', 'nothing', 'nobody', 'neither', 'nowhere', 'hardly', 'scarcely',
                  'barely', 'doesn’t', 'isn’t', 'wasn’t', 'shouldn’t', 'wouldn’t', 'couldn’t', 'won’t', "can't",
                  "don't"]

INTERROGATIVE_WORDS = ['what', 'who', 'when', 'where', 'which', 'why', 'how']

TENTATIVE_WORDS = ['appears to', 'can', 'could', 'it is likely', 'it is possible', 'it is probable', 'it is unlikely',
                   'may', 'might', 'possibly', 'probably', 'seems to', 'suggests that', 'tends to']

REASON_WORDS = ['because', 'reason', 'as a result', 'since', 'therefore']


@lru_cache(1)
def onegram_count() -> pd.DataFrame:
    """
    Get counts of 1-gram from Peter Norvig's list.

    Returns:
        DataFrame with one-gram, count and idf scores.
    """
    df = pd.read_csv('https://norvig.com/ngrams/count_1w.txt',
                     sep='\t',
                     header=None,
                     names=['word', 'count'])
    df['idf'] = np.log(df['count'].sum() / df['count'])
    df.sort_values(by='idf', ascending=True, inplace=True)
    return df



import hashlib
from typing import List


def md5_hash(text: str) -> str:
    """
    Generate MD5 hash of a text.

    Args:
        text: String

    Returns:
        MD5 hash
    """
    return hashlib.md5(text.encode('utf-8')).hexdigest()


def sha256hash(text: str) -> str:
    """
    Generate MD5 hash of a text.

    Args:
        text: String

    Returns:
        SHA256 hash
    """
    return hashlib.sha256(text.encode('utf-8')).hexdigest()


def window(tokens,
           size: int = 3):
    """
    Generate samples for a window size.

    Example:
    ```python
    >>> window(['a', 'b', 'c', 'd'], size=2)
    [(['a', 'b'], 'c'), (['b', 'c'], 'd')]
    ```

    Args:
        tokens: List of tokens
        size: Window size

    Returns:
        List of windowed samples
    """
    return [(tokens[i: i + size], tokens[i + size])
            for i in range(0, len(tokens) - size, 1)]


def offset_by_one(x,
                  sequence_length: int = 3):
    """
    Generate a list of small sequences offset by 1.

    Usage:

    ```python
    >>> offset_by_one([1, 2, 3, 4, 5], sequence_length=3)
    [([1, 2, 3], [2, 3, 4])]
    ```

    Args:
        x: Python list
        sequence_length: Chunk size

    Returns:

    """
    sl = sequence_length
    return [(x[i:i + sl], x[i + 1:i + sl + 1])
            for i in range(0, len(x) - sl - 1, sl)]


def num_words(text: str) -> int:
    """
    Counts the number of words using whitespace as delimiter.

    Args:
        text: Sentence

    Returns:
        Number of words
    """
    return len(text.split())


def unique_chars(texts: List[str]) -> List[str]:
    """
    Get a list of unique characters from list of text.

    Args:
        texts: List of sentences

    Returns:
        A sorted list of unique characters
    """
    return sorted(set(''.join(texts)))


def is_non_ascii(text: str) -> bool:
    """
    Check if text has non-ascci characters.

    Useful heuristic to find text containing emojis and non-english
    characters.

    Args:
        text: Sentence

    Returns:
        True if the text contains non-ascii characters.
    """
    try:
        text.encode('ascii')
        return False
    except UnicodeEncodeError:
        return True



from typing import List, Union

from sklearn.cluster import KMeans
from sklearn.decomposition import TruncatedSVD
from sklearn.feature_extraction.text import TfidfVectorizer, CountVectorizer

from .metrics import n_clusters
from sklearn.pipeline import Pipeline
import pandas as pd


def cluster_text(texts: List[str],
                 embedding: str = 'tf-idf',
                 return_dataframe: bool = True) -> Union[pd.DataFrame, List[int]]:
    """
    Quickly cluster a list of sentences for EDA.

    Args:
        texts: List of sentences
        embedding: 'tf-idf' or 'count'
        return_dataframe: Whether to return as dataframe or a list of cluster labels

    Returns:

    """
    n = n_clusters(texts)
    if embedding == 'tf-idf':
        vectorizer = TfidfVectorizer(strip_accents='ascii',
                                     stop_words='english',
                                     sublinear_tf=True)
    elif embedding == 'count':
        vectorizer = CountVectorizer(strip_accents='ascii',
                                     stop_words='english')
    else:
        raise Exception('Invalid argument for embedding')

    cluster_pipeline = Pipeline([
        ('vectorizer', vectorizer),
        ('pca', TruncatedSVD()),
        ('kmeans', KMeans(n))
    ])
    clusters = cluster_pipeline.fit_predict(texts)
    if return_dataframe:
        return pd.DataFrame({'text': texts, 'cluster': clusters})
    else:
        return clusters
